// $Id: FreeEnergyModel.h 1354 2023-06-15 00:03:41Z ge $
/// \file FreeEnergyModel.h
/// \brief contains the FreeEnergyModel class
///
/// $Revision: 1354 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_FREEENERGYMODEL_H
#define GBC_FREEENERGYMODEL_H "$Id: FreeEnergyModel.h 1354 2023-06-15 00:03:41Z ge $"
#include <algorithm>
#include "SequenceInfo.h"
#include "SequenceDataset.h"
#include "IndividualNearestNeighbours.h"

namespace gbc
{
template<class _Model=IndividualNearestNeighbours>
// FreeEnergyModel combines a model like IndividualNearestNeighbours with a data set
class FreeEnergyModel
  {
  public:
    typedef _Model                                 model_type;
    typedef typename model_type::value_type        value_type;
    typedef std::valarray<value_type>              vector_type;
    typedef typename model_type::sequenceinfo_type sequenceinfo_type;
    typedef typename model_type::matrix_type       matrix_type;
    typedef typename model_type::svd_deque_type    svd_deque_type;
    typedef SequenceDataset<sequenceinfo_type>     sequence_dataset_type;
    typedef std::deque<std::string>                string_deque_type;

    sequence_dataset_type dataset;
    model_type            *model;
    std::string           data_set_type;
    svd_deque_type        SVDVariables;

    FreeEnergyModel(void): model() {}

   inline void verify(std::ostream& out)
      {
      dataset.calculate_and_print_verify(out);
      }

   inline void print_tex(std::ostream& out)
     {
     dataset.print_tex(out);
     }

    /// \brief Calculates the difference between the predicted and the adjusted temperature
    inline void predict_all(void)
      {
      typename sequence_dataset_type::sequence_info_deque_type::iterator si=dataset.Raw_dataset.begin();
      for(si=dataset.Raw_dataset.begin(); si != dataset.Raw_dataset.end(); si++)
        {
        model->Sequence_info=&(*si);
        model->reset_calculation();
        model->melting_temperature();
        }
      }
      
    inline void predict_and_print_sequentially(std::ostream& out)
      {
      bool first_line=true;
      std::ifstream xml;
      double multspconc=1;
      typename sequence_dataset_type::species_concentration_deque_type spconc;
      size_t number_of_sequences=0;
      for (auto xmlfile : dataset.Data_file_deque)
        {
        dataset.open_and_read_xml_header(xml,xmlfile);
        while(!xml.eof()) 
          {
          dataset.read_xml_line(xml,xmlfile,spconc,multspconc);
          number_of_sequences++;
          for(auto &si : dataset.Raw_dataset) 
            {
            predict_all();
          
            dataset.collect_tm_difference(si);
          
            if (first_line) si.print_head(out);
            first_line=false;
            si.print(out);
            }
          dataset.Raw_dataset.clear();
          }
        }
      dataset.calculate_tm_statistics();
      }
      
      
    //Rules are read from parameter files, make sure that this is set before read_xml
    //otherwise the rules are not applied to the dataset
    void read_rules_and_parameters_from_list_of_files(string_deque_type &par_file_deque)
      {
      dataset.read_rules_from_list_of_files(par_file_deque);
      model->get_parameters_from_list_of_files(par_file_deque);
      }

    
    inline void set_svd_variables(const svd_deque_type &svd)
      {
      SVDVariables=svd;
      }
      
    inline void set_svd_matrices(matrix_type &A, matrix_type &B)
      {
      matrix_type M;
      int i=0;
      for(auto &si : dataset.Raw_dataset)
        {
        model->Sequence_info=&si;
        value_type b;
        model->get_svd_line(SVDVariables,M,b);
        if (A.Rows ==0 or A.Columns == 0) A.resize(dataset.Raw_dataset.size(),M.Columns);
        for (int j=0; j < M.Columns; j++) A(i,j)=M(0,j);
        if (B.Rows ==0 or B.Columns == 0) B.resize(dataset.Raw_dataset.size(),1);
        B(i,0)=b;
        i++;
        }
      CERR_DEBUG(DFEM_SHSVDMAT) << "A(" << A.Rows << "," << A.Columns << ")" << std::endl;
      CERR_DEBUG(DFEM_SHSVDMAT) << A << std::endl;
      CERR_DEBUG(DFEM_SHSVDMAT) << "B(" << B.Rows << "," << B.Columns << ")" << std::endl;
      CERR_DEBUG(DFEM_SHSVDMAT) << B << std::endl;
      }
  
    inline void adjust_svd_units(vector_type &x)
      {
      int i=0;
      for(auto &v: SVDVariables)
        {
        if (reg_match(v,":enthalpy$")) x[i] /= 1000;//cal/mol => kcal/mol
        i++;
        }
      }

  };

}
#endif
